class NiederreiterCryptosystem:

	def __init__(self, n, m, g):
		#Construct the Goppa code
		goppa_code = GoppaCode(n,m,g);
		assert goppa_code.generator_matrix().nrows() <> 0, "Generator Matrix is empty.";
		k = goppa_code.generator_matrix().nrows();

		# Set up the random scrambler matrix
		S = matrix(GF(2),n-k,[random()<0.5 for _ in range((n-k)^2)]);
		while (rank(S) < n-k) :
			S[floor((n-k)*random()),floor((n-k)*random())] +=1;

		# Set up the permutation matrix
		rng = range(n); P = matrix(GF(2),n);
		for i in range(n):
			p = floor(len(rng)*random());
			P[i,rng[p]] = 1; rng=rng[:p]+rng[p+1:];

		#Remember these values
		self._m_GoppaCode = goppa_code;
		self._g = g;
		self._t = g.degree();
		self._S = S;
		self._P = P;
		self._PublicKey = S*(self._m_GoppaCode.parity_check_matrix())*P;

	#This is a help function which will be useful for encryption.
	def _GetRowVectorWeight(self,n):
		weight = 0;
		for i in range(n.ncols()):
			if n[0,i] == 1:
				weight = weight+1;
		return weight;

				
	def Encrypt(self, message):
	#Assert that the message is of the right length
		assert (message.ncols() == self.public_key().ncols()), "Message is not of the correct length."; 
		assert (self._GetRowVectorWeight(message) <= self.max_num_errors()), "Message weighs at most %d." %(self.max_num_errors());
		
		code_word = self.public_key()*(message.transpose());
		return code_word;
	
	def Decrypt(self, received_word):
		assert (received_word.nrows() == self.public_key().nrows()), "Received word is not of the correct length.";
                assert (received_word.ncols() == 1), "Received word is not of the correct length."
		#Strip off the random scrambler matrix and decode the received word.
		message = ~(self._S)*received_word;
		
		#Convert message into syndrome polynomial.
		t = self.max_num_errors();
		m = message.nrows()/t;
		g = self.goppa_code().goppa_polynomial();		
		F2 = GF(2);
		F_2m = g.base_ring(); Z = F_2m.gen();
		PR_F_2m = g.parent(); X = PR_F_2m.gen();		
		syndrome_poly = 0;
                
		for i in range(t):
                        tmp = [];
			for j in range(m):
				tmp.append(message[i*m+j,0]);
                        syndrome_poly += F_2m(tmp[::-1])*X^i;
                
		#Retrieve using syndrome decoding algorithm. 		
		message = self.goppa_code().SyndromeDecode(syndrome_poly);
		
		#Solve the system to determine the original message.
		message = self._P.transpose()*message.transpose();
		return message;			

	
	def _str2num(self,ascii_text):
		#Turn ascii text into big integers by treating the ASCII values 
		#of each character as digits in base 256:
    		return ZZ(map(ord,ascii_text),256);

	def _common_primroot(self,p,q):
	    # Finds a primitive root common to both p and q
	    ps = [f for f,e in factor(p-1)]
	    qs = [f for f,e in factor(q-1)]
	    a=2
	    while true:
		outp = true
		for d in ps:
		    if power_mod(a,(p-1)//d,p)==1:
		        outp=false
		outq =  true
		for d in qs:
		    if power_mod(a,(q-1)//d,q)==1:
		        outp=false
		if outp and outq:
		    return a
		    break
		else:
		    a+=1 

	def CryptoHash(self, plaintext, hash_value_length):
		#This cryptographic hash function was proposed by Adi Shamir.
		#For more details, visit http://amca01.wordpress.com/2010/09/03/a-quick-and-dirty-hash-function/
		# Implements Shamir's hash function h = a^m (mod p*q) where a is a
    		# number of high multiplicative order mod p*q.  This works well if
    		# a is chosen to be a common primitive root of both p and q
		p = next_prime(randint(2^150,2^151));
		q = next_prime(randint(2^150,2^151));
		a = self._common_primroot(p,q);
		
		hashvalue = power_mod(a,self._str2num(plaintext),p*q)%(2^hash_value_length);
		hashvalue = bin(hashvalue).zfill(hash_value_length);
    		return hashvalue[2:];
		
	def Signature(self, message):
		#message lenght: m*t;  weight: t.
		return message;
		
	
	def Verify(self, signature):	
		return signature;	

	#Accessors
	def public_key(self):
			return self._PublicKey;

	def goppa_code(self):
			return self._m_GoppaCode;

	def max_num_errors(self):
			return self._t;
		
